// SerialMon, firmware for serial communications monitor
// Copyright (C) 2008, Frank Tkalcevic, www.franksworkshop.com

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include "Common.h"
#include "bootloader.h"

#define XMODEM_DATA_BUFFER_SIZE		128
#define XMODEM_SOH					0x01 // Start of header
#define XMODEM_EOT					0x04 // End of transmission
#define XMODEM_ACK					0x06 // Acknowledge
#define XMODEM_NACK					0x15 // Not acknowledge
#define XMODEM_RRC					'C' // Ready to receive with CRC

#define BOOTLOADER_ADDR				0x1800

// ### SWAP ### Swap Nibbles Of Register (reg(0...3) = reg(4...7), reg(4...7) = reg(0...3)) 
#define SWAP(reg) asm volatile ("swap %0" : "=r" (reg) : "0" (reg)) 

typedef struct 
{ 
	uint8_t Bytes[2]; 
} ByteArray; 

#define BYTE(x,y)                 ((ByteArray*)&x)->Bytes[y] 

static BOOTLOADER_SECTION byte get_byte( void )
{
	byte nTimeout = 0;
	while ( !( UCSR0A & _BV(RXC0) ) )
	{
		if ( TIFR1 & _BV(OCF1A) )
		{
			// tick
			TIFR1 = _BV(OCF1A);
			nTimeout++;
			if ( nTimeout == 120 )	// 120 seconds, 2 minutes
			{
				run_bootloader();	// if we timeout waiting for another byte, there is no use
									// restarting as the flash is probably stuffed.  Restart bootloader.
			}
		}
	}
	return UDR0;
}

static BOOTLOADER_SECTION void put_byte( byte c )
{
	while ( !( UCSR0A & _BV(UDRE0) ) )
		;
	UDR0 = c;
}


static BOOTLOADER_SECTION void boot_program_page( uint32_t page, uint8_t *buf )
{
	uint16_t i;

	eeprom_busy_wait ();
	boot_page_erase (page);
	boot_spm_busy_wait (); // Wait until the memory is erased.
	for (i=0; i<SPM_PAGESIZE; i+=2)
	{
		// Set up little-endian word.
		uint16_t w = *buf++;
		w += (*buf++) << 8;
		boot_page_fill (page + i, w);
	}
	boot_page_write (page); // Store buffer in flash page.
	boot_spm_busy_wait(); // Wait until the memory is written.

	// Reenable RWW-section again. We need this if we want to jump back
	// to the application after bootloading.
	boot_rww_enable ();
}


static BOOTLOADER_SECTION bool crc_check( byte *ptr, byte crchi, byte crclo )
{
	byte count = 128;
	uint16_t crc=0;
	while (count-- > 0)
	{
		crc = crc ^ (int) *ptr++ << 8;
		char i = 8;
		do
		{
			if (crc & 0x8000)
			{
				crc = crc << 1 ^ 0x1021;
			}
			else
			{
				crc = crc << 1;
			}
		} while(--i);
	}

	if ( ((crc >> 8) == crchi) && ((crc & 0xff) == crclo) )
		return true;
	else
		return false;
}


static BOOTLOADER_SECTION byte FromHex( char c )
{
	if ( c <= '9' )
		return c - '0';
	else if ( c <= 'F' )
		return c - 'A' + 0xA;	
	else
		return c - 'a' + 0xA;	
} 



void run_bootloader( void )  __attribute__ ((naked))  __attribute__ ((section (".bootloader")));
void run_bootloader( void )
{
	byte xmodem_buffer[XMODEM_DATA_BUFFER_SIZE];
	byte program_buffer[SPM_PAGESIZE];
	byte nProg = 0;
	uint16_t nPage = 0;
	byte nPairs;
	uint16_t nAddress;
	enum EIntelHexProcessingStates 
	{
		START,
		READ_DATA_PAIRS,
		READ_ADDR1,
		READ_ADDR2,
		READ_DATA_TYPE,
		READ_DATA,
		DONE
	} state = START;
	bool bFirstNibble = true;
	byte temp_byte;

	cli();

	// configure timer1 to OC every 1 sec
	TCCR1A = 0;
	TCCR1B = _BV(WGM12) | _BV(CS12) | _BV(CS10);
	OCR1A = 7812;
	TCNT1 = 0;
	TIFR1 = _BV(OCF1A);


	// send 'C" every 3 seconds till we start.
	put_byte( XMODEM_RRC );
	byte n = 0, clock=0;
	for (;;)
	{
		if ( TIFR1 & _BV(OCF1A) )
		{
			// tick
			TIFR1 = _BV(OCF1A);
			n++;
			if ( n == 3 )
			{
				put_byte( XMODEM_RRC );
				n=0;
				clock++;
				if ( clock == 40 )	// 40 * 3 seconds
					goto Restart;
			}
		}
		if ( UCSR0A & _BV(RXC0) )
			if ( get_byte() == XMODEM_SOH )
				break;
	}

	// read xmodem packet of HEX data
	byte packet = 1;
	for (;; )
	{
		n = get_byte();	// packet #
		if ( n != packet )
		{
			//DebugSendString( "Wrong Packet\r\n" );
		}
		n = get_byte();
		if ( n != 255 - packet )		// ~packet #
		{
			//DebugSendString( "Wrong Not Packet\r\n" );
			// error
		}


		// data
		for ( n = 0; n < XMODEM_DATA_BUFFER_SIZE; n++ )
		{
			xmodem_buffer[n] = get_byte();
		}
		byte crchi = get_byte();
		byte crclo = get_byte();

		if ( !crc_check( xmodem_buffer, crchi, crclo ) )
		{
			// error
			//DebugSendString( "CRC failed\r\n" );
			put_byte( XMODEM_NACK );
		}
		else
		{
			// Parse the .hex file

			// <:><DataPairs><AddressHi><AddressLo><DataType><Data>...<CheckSum>

			for ( n = 0; n < XMODEM_DATA_BUFFER_SIZE; n++ )
			{
				char c = xmodem_buffer[n];

				if ( state != START )
				{
					if ( bFirstNibble )
					{
						temp_byte = FromHex(c);
						SWAP( temp_byte );
					}
					else
					{
						temp_byte |= FromHex(c);
					}
					bFirstNibble = !bFirstNibble;
					if ( !bFirstNibble )
						continue;
				}

				if ( state == START )
				{
					if ( c == ':' )	// start of line
						state = READ_DATA_PAIRS;
				}
				else if ( state == READ_DATA_PAIRS )
				{
					nPairs = temp_byte;
					state = READ_ADDR1;
				}
				else if ( state == READ_ADDR1 )
				{
					BYTE(nAddress,1) = temp_byte;
					state = READ_ADDR2;
				}
				else if ( state == READ_ADDR2 )
				{
					BYTE(nAddress,0) = temp_byte;
					// if addr >= BOOTLOADER ADDR, store what we have, stop storing the rest - we don't overwrite the bootloader
					if ( nAddress >= BOOTLOADER_ADDR )
					{
						if ( nProg > 0 )
							boot_program_page( nPage, program_buffer );
						state = DONE;
					}
					else
						state = READ_DATA_TYPE;
				}
				else if ( state == READ_DATA_TYPE )
				{
					if ( temp_byte != 0 )
					{
						// EOF, store what we have, stop storing the rest
						if ( nProg > 0 )
							boot_program_page( nPage, program_buffer );
						state = DONE;
					}
					else
					{
						// next packet
						state = READ_DATA;
					}
				}
				else if ( state == READ_DATA )
				{
					program_buffer[nProg++] = temp_byte;
					if ( nProg >= SPM_PAGESIZE )
					{
						// store program memory
						boot_program_page( nPage, program_buffer );
						nPage += SPM_PAGESIZE;
						nProg = 0;
					}
					nPairs--;
					if ( nPairs == 0 )
						state = START;
				}
			}

			put_byte( XMODEM_ACK );
			packet++;
		}

		n = get_byte();
		if ( n == XMODEM_EOT )
		{
			put_byte( XMODEM_ACK );
			break;
		}
		else if ( n == XMODEM_SOH )
			continue;
		else
			break;
	}

Restart:
	wdt_enable( WDTO_500MS );
	for (;;) ;
	//void (*jump_to_app)(void) = 0x0000;
	//jump_to_app();

	//// 
	//for (;;) 
	//{
	//	if ( TIFR0 & _BV(OCF0A) )
	//	{
	//		TIFR0 = _BV(OCF0A);
	//		n++;
	//		if (n==200)
	//		{
	//			n = 0;
	//			if ( PORTB & _BV(PB7) )
	//				PORTB &= ~_BV(PB7);
	//			else
	//				PORTB |= _BV(PB7);
	//		}
	//	}
	//	continue;
	//}
}
