// SerialMon, firmware for serial communications monitor
// Copyright (C) 2008, Frank Tkalcevic, www.franksworkshop.com

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include "Common.h"
#include "serial.h"
#include "bootloader.h"

//#define DEBUG

#ifdef DEBUG
#endif


// ### SWAP ### Swap Nibbles Of Register (reg(0...3) = reg(4...7), reg(4...7) = reg(0...3)) 
#define SWAP(reg) asm volatile ("swap %0" : "=r" (reg) : "0" (reg)) 


#define LED_LINE0			PD5
#define LED_LINE1			PD6
#define LED_PORT			PORTD

#define LED0(b)				if ( b ) LED_PORT |= _BV(LED_LINE0); else LED_PORT &= ~_BV(LED_LINE0);
#define LED1(b)				if ( b ) LED_PORT |= _BV(LED_LINE1); else LED_PORT &= ~_BV(LED_LINE1);

#define IN_LINE0			PD2		// PCINT18, INT0
#define IN_LINE1			PD3		// PCINT19, INT1
#define IN_PIN				PIND
#define IN_LINE0_PCINT		PCINT18
#define IN_LINE1_PCINT		PCINT19


static volatile byte nTick;
static bool bRunning;
enum EIncomingMsgType
{
	ConfigLine0 = 0x02,
	ConfigLine1 = 0x03,
	AutoConfigLine0 = 0x04,
	AutoConfigLine1 = 0x05,
	Start = 0x06,
	Stop = 0x08
};

struct SInMsgHdr
{
	uint8_t nStx;
	uint8_t nLen;
	uint8_t nCmd;
};

struct SLineConfig
{
	struct SInMsgHdr hdr;
	uint32_t nBaud;
	uint8_t nDataBits;
	uint8_t nParity;
	uint8_t nStopBits;
};

#define MAX_MSG_LEN		(sizeof(struct SLineConfig))
#define MSG_START		0xAA

enum ERS232ReadState
{
	WaitForStart,
	WaitForMsg
};


struct SLine			// Working data for each line
{
	uint32_t Baud;
	byte DataBits;
	byte StopBits;
	byte Parity;
	byte TCCRnB;		// TCCRnB contents - contains prescaler setting
	byte OCRnA;			// contains OCRnA value to trigger OCRnA 
	byte BitsToRead;	// How many bits to read - start + data + parity + stop -1
	byte Bit;			// Which bit are we reading
	byte Data;			// The data we build up as we read it
	byte Mask;			// The bit mask to build the bit
	//byte Ninth;			// to store the ninth data bit
	byte t;
};

static struct SLine Line[2];
static volatile byte nByteCount;

/*
Outputs
    X X X X X X X X  XXXXXXXX
    ^ ^ ^ ^ ^ ^ ^ ^
    | | | | | | | |
    | | | | | | | +---  9th Bit
    | | | | | | +-----  Line 0/Line 1
    | | | | + +-------  Data Type 0 - data, 1 Error, 2 Timing, 3 Autobaud Complete
*/
#define TYPE_DATA		(0<<2)
#define TYPE_ERROR		(1<<2)
#define TYPE_TIMING		(2<<2)
#define TYPE_AUTOBAUD	(3<<2)

#define LINE_ID_0		(0<<1)
#define LINE_ID_1		(1<<1)
	


/************************************************************************
			Interrupts
*************************************************************************/

SIGNAL (SIG_OUTPUT_COMPARE1A)	// Timer1 output compare interrupt - ms timer
{
	sei();

	if ( nByteCount == 16 )
	{
		if ( bRunning )
		{
			RS232SendChars( TYPE_TIMING, 16 );
		}
		nByteCount = 0;
	}
	else
	{
		nByteCount++;
	}

	static byte o = 0;
	o++;
	if (o==10)
	{
		o = 0;
		LED0(false);
		LED1(false);
	}
}

/*****************************
		Serial Line 0
******************************/

SIGNAL (SIG_INTERRUPT0)		// Serial Line 0
{
	// set up the timer to interrupt us in the middle of each bit.  The first interrupt must be in 1/2 a baud
	TCNT0 = 0;
	TCCR0A = _BV(WGM01);
	TCCR0B = Line[0].TCCRnB;
	OCR0A = Line[0].OCRnA/2;

	// We don't need the interrupt anymore
	EIMSK &= ~_BV(INT0);

	// Set up the timer interrupt
	TIFR0 |= _BV(OCF0A);
	TIMSK0 = _BV(OCIE0A);
	Line[0].Bit = 0;
	Line[0].Mask = 1;
	Line[0].Data = 0;
	Line[0].t = nByteCount;
	nByteCount = 0;

	LED0(true);
}

SIGNAL (SIG_OUTPUT_COMPARE0A)	// Timer0 output compare interrupt - Line 0 baud timer
{
	// Set OCR0A to full bit timing
	OCR0A = Line[0].OCRnA;

	byte bit;
	if ( (byte)IN_PIN & (byte)_BV(IN_LINE0) )
		bit = (byte)1;
	else
		bit = (byte)0;
	sei();

	if ( Line[0].Bit == 0 )	// Start bit
	{
		Line[0].Bit++;
	}
	else if ( Line[0].Bit <= Line[0].DataBits )	// Past the start bit
	{
		Line[0].Bit++;
		//if ( Line[0].Bit == 9 )	// Special for 9 bit data
		//	Line[0].Ninth = bit;
		//else
		{
			if ( bit != (byte)0 )
				Line[0].Data |= Line[0].Mask;
			Line[0].Mask <<= 1;
		}

		if ( Line[0].Bit == Line[0].BitsToRead )	// Got to the stop bit
		{
			if ( bit )
			{
				// stop bit should be low.  This is a framing error, or we just haven't resynced yet.
			}


			cli();
			RS232SendChars( (Line[0].t << 4) | TYPE_DATA | LINE_ID_0 /*| Line[0].Ninth*/, Line[0].Data );
			TCNT1 = 0;
			sei();
			nByteCount = 0;

			// Wait for the next start edge
			EIFR |= _BV(INTF0);
			EIMSK |= _BV(INT0);
			TIMSK0 = 0;

			return;
		}
	}
}

/*****************************
		Serial Line 1
******************************/

SIGNAL (SIG_INTERRUPT1)		// Serial Line 1
{
	// set up the timer to interrupt us in the middle of each bit.  The first interrupt must be in 1/2 a baud
	TCNT2 = 0;
	TCCR2A = _BV(WGM21);
	TCCR2B = Line[1].TCCRnB;
	OCR2A = Line[1].OCRnA/2;

	// We don't need the interrupt anymore
	EIMSK &= ~_BV(INT1);

	// Set up the timer interrupt
	TIFR2 |= _BV(OCF2A);
	TIMSK2 = _BV(OCIE2A);
	Line[1].Bit = 0;
	Line[1].Mask = 1;
	Line[1].Data = 0;
	Line[1].t = nByteCount;
	nByteCount = 0;

	LED1(true);
}


SIGNAL (SIG_OUTPUT_COMPARE2A)	// Timer2 output compare interrupt - Line 2 baud timer
{
	// Set OCR0A to full bit timing
	OCR2A = Line[1].OCRnA;

	byte bit;
	if ( IN_PIN & _BV(IN_LINE1) )
		bit = 1;
	else
		bit = 0;
	sei();

	if ( Line[1].Bit == 0 )	// Start bit
	{
		Line[1].Bit++;
	}
	else if ( Line[1].Bit <= Line[1].DataBits )	// Past the start bit
	{
		//if ( Line[1].Bit == 9 )	// Special for 9 bit data
		//	Line[1].Ninth = bit;
		//else
		{
			Line[1].Bit++;
			if ( bit )
				Line[1].Data |= Line[1].Mask;
			Line[1].Mask <<= 1;
		}
		if ( Line[1].Bit == Line[1].BitsToRead )	// Got to the stop bit
		{
			if ( bit )
			{
				// stop bit should be low.  This is a framing error, or we just haven't resynced yet.
			}


			cli();
			RS232SendChars( (Line[1].t << 4) | TYPE_DATA | LINE_ID_1 /*| Line[1].Ninth*/, Line[1].Data );
			TCNT1 = 0;
			sei();

			nByteCount = 0;

			// Wait for the next start edge
			EIFR |= _BV(INTF1);
			EIMSK |= _BV(INT1);
			TIMSK2 = 0;

			return;
		}
	}
}


#define AUTOBAUD_BITS_TO_SAMPLE		60

static volatile uint16_t nLastBitPeriod;
static volatile bool bAutoBaudDone;
static bool bFirstAutoBaud;
static byte nAutoBaudCount;
static byte nAutoBaudLine;

SIGNAL (SIG_PIN_CHANGE2)	// Used to time autobauding
{
	uint16_t t = TCNT1;
	TCNT1 = 0;

	if ( bFirstAutoBaud )
	{
		bFirstAutoBaud = false;
	}
	else
	{
		if ( TIFR1 & _BV(TOV1) )	
		{
			// An overflow.  We ignore this bit and try again.
		}
		else
		{
			if ( t < nLastBitPeriod )
				nLastBitPeriod = t;
			nAutoBaudCount--;
			if ( nAutoBaudCount == 0 )
			{
				// Done.
				PCICR = 0;
				bAutoBaudDone = true;
			}
		}

	}
	TIFR1 = _BV(TOV1);	// turn off overflow.
}


static void SetClockTimer( void )		// byte timer
{
	// set the timer to overflow after every data byte (start+data+parity+stop) of the fastest baud.
	uint16_t nPeriod = 0xFFFF;
	for ( int line = 0; line < 2; line++)
	{
		int nBits = Line[line].BitsToRead + 1;		// +1 for start bit
		uint32_t nTicks = CPU_CLK;
		nTicks *= nBits;
		nTicks /= Line[line].Baud;
		uint16_t _nTicks = nTicks;
		if ( _nTicks < nPeriod )
			nPeriod = _nTicks;
	}

	struct SPrescaling
	{
		byte nMask;
		byte nShift;
	};
	// Timer 1 prescaling information 
	static struct SPrescaling PreScale1[] = 
	{ 
		{                         _BV(CS10), 0 },
		{             _BV(CS11)            , 3 },
		{             _BV(CS11) | _BV(CS10), 6 },
		{ _BV(CS12)                        , 8 },
		{ _BV(CS12)             | _BV(CS10), 10 } };

	for ( int i = 0; i < countof(PreScale1); i++ )
	{
		uint32_t nCount = nPeriod >> PreScale1[i].nShift;
		if ( nCount < 0x10000 )
		{
			TCCR1A = 0;	// CTC mode top=ocr1a
			TCCR1B = _BV(WGM12) | PreScale1[i].nMask;
			OCR1A = nCount;
			TIMSK1 |= _BV(OCIE1A);
			break;
		}
	}
}

static void ioinit (void)
{
	///////////////////////
	// Set IO registers	
	///////////////////////
	DDRD = OUTPUT(LED_LINE0) | OUTPUT(LED_LINE1) | INPUT(IN_LINE0) | INPUT(IN_LINE1);
	PORTD = 0;

	// ms timer. @ 20MHz
	SetClockTimer();
}


static void SetPrescaler( byte nLine, uint32_t nPeriod )
{
	struct SPrescaling
	{
		byte nMask;
		byte nShift;
	};
	// Timer 0 prescaling information for line 0
	static struct SPrescaling PreScale0[] = 
	{ 
		{                         _BV(CS00), 0 },
		{             _BV(CS01)            , 3 },
		{             _BV(CS01) | _BV(CS00), 6 },
		{ _BV(CS02)                        , 8 },
		{ _BV(CS02)             | _BV(CS00), 10 }, };
	// Timer 2 prescaling information for line 1
	static struct SPrescaling PreScale2[] = 
	{ 
		{                         _BV(CS20), 0 },
		{             _BV(CS21)            , 3 },
		{             _BV(CS21) | _BV(CS20), 5 },
		{ _BV(CS22)                        , 6 },
		{ _BV(CS22)             | _BV(CS20), 7 },
		{ _BV(CS22) | _BV(CS21)            , 8 },
		{ _BV(CS22) | _BV(CS21) | _BV(CS20), 10 }, };

	if ( nLine == 0 )
	{
		for ( int i = 0; i < countof(PreScale0); i++ )
		{
			uint32_t nCount = nPeriod >> PreScale0[i].nShift;
			if ( nCount < 256 )
			{
				Line[nLine].TCCRnB = PreScale0[i].nMask;
				Line[nLine].OCRnA = nCount;
				break;
			}
		}
	}											  
	if ( nLine == 1 )
	{
		for ( int i = 0; i < countof(PreScale2); i++ )
		{
			uint32_t nCount = nPeriod >> PreScale2[i].nShift;
			if ( nCount < 256 )
			{
				Line[nLine].TCCRnB = PreScale2[i].nMask;
				Line[nLine].OCRnA = nCount;
				break;
			}
		}
	}
}
												    


static void ConfigureLine( byte nLine, uint32_t nBaud, byte nDataBits, byte nParity, byte nStopBits )
{
	Line[nLine].Baud = nBaud;
	Line[nLine].DataBits = nDataBits;
	Line[nLine].Parity = nParity;
	Line[nLine].StopBits = nStopBits;
	Line[nLine].BitsToRead = nDataBits;
	Line[nLine].BitsToRead += (nStopBits==20 ? 2 : 1);		// 1.5 currently not supported
	Line[nLine].BitsToRead += (nParity? 1 : 0);

	// Calculate the timer prescaler and counter
	uint32_t nTicks = CPU_CLK;
	nTicks /= nBaud;

	SetPrescaler( nLine, nTicks );
	SetClockTimer();
}


static void StartListening( void )
{
	EICRA = _BV(ISC11) | _BV(ISC01);	// INT0,INT1 will interrupt on falling edge - hopefully a start bit.
	EIMSK = _BV(INT1) | _BV(INT0);
	bRunning = true;
}

static void StopListening( void )
{
	bRunning = false;
	EIMSK = 0;
	TIMSK0 = 0;
	TIMSK2 = 0;
	LED0(false);
	LED1(false);
}

void AutoConfigLine( byte nLine )
{
	StopListening();	// Can't auto config while other stuff is happening

	// Set timer 1, 16 bit timer, so we can count the samples between bits.
	// we only use this as a counter, so it doesn't interrupt
	TCCR1A = 0;
	TCCR1B = _BV(CS10);		// clk/1
	TCCR1C = 0;
	TCNT1 = 0;

	nLastBitPeriod = 0xFFFF;
	bAutoBaudDone = false;
	bFirstAutoBaud = true;
	nAutoBaudCount = AUTOBAUD_BITS_TO_SAMPLE;
	nAutoBaudLine = nLine;

	// Use PCINT to watch for pin changes.
	if ( nLine == 0 )
		PCMSK2 = _BV(IN_LINE0_PCINT);
	else
		PCMSK2 = _BV(IN_LINE1_PCINT);

	PCICR = _BV(PCIE2);
}

static void ProcessMessage( struct SInMsgHdr *cmd )
{
	switch ( cmd->nCmd )
	{
		case ConfigLine0:
		case ConfigLine1:
			{
				struct SLineConfig *config = (struct SLineConfig *)cmd;
				byte nLine = config->hdr.nCmd & 1;
				ConfigureLine( nLine, config->nBaud, config->nDataBits, config->nParity, config->nStopBits );
			}
			break;
		case AutoConfigLine0:
		case AutoConfigLine1:
			AutoConfigLine( cmd->nCmd & 1 );
			break;
		case Start:
			StartListening();
			break;
		case Stop:
			StopListening();
			break;
	}

}


static byte cmd_buf[MAX_MSG_LEN];
static byte buf_ptr;
static byte nRS232ReadState;

static void ProcessComms( void )
{
	byte b = RS232ReadByte();

	switch ( nRS232ReadState )
	{
		case WaitForStart:
			if ( b != MSG_START )
				break;
			buf_ptr = 0;
			nRS232ReadState = WaitForMsg;
			// Fall through

		case WaitForMsg:
			cmd_buf[buf_ptr] = b;
			buf_ptr++;
			if ( buf_ptr >= sizeof(struct SInMsgHdr) )
			{
				if ( buf_ptr >= ((struct SInMsgHdr *)cmd_buf)->nLen )
				{
					ProcessMessage( (struct SInMsgHdr *)cmd_buf );
					nRS232ReadState = WaitForStart;
				}
				else if ( buf_ptr >= countof(cmd_buf))	// stop buffer overflow
					nRS232ReadState = WaitForStart;
			}
			break;
	} 
}

int main (void)
{
	byte reason = MCUSR;
	MCUSR = 0;

	ioinit();

	InitRS232( 9 );			// 9 = 125kHz@20MHz
	UCSR0A |= _BV(U2X0);	// x2 - 250kHz


	nRS232ReadState = WaitForStart;
	nByteCount = 0;
	bRunning = false;

	sei();

	RS232Send_P( PSTR("Serial Monitor v1.0 Initialised.\r\n") );
	if ( reason != 0 )
	{
		RS232Send_P( PSTR("Last restart reason = "));
		RS232SendInt( reason );
		RS232Send_P( PSTR(".\r\n" ) );
	}

	ConfigureLine( 0, 38400, 8, 0, 10 );	// defaults
	ConfigureLine( 1, 38400, 8, 0, 10 );
	//AutoConfigLine( 0 );
	//StartListening();

	byte nOldTick = nTick;
	for (;;)
	{
		if ( nOldTick != nTick )
		{
			//RS232Send_P( PSTR("Tick\r\n"));
			nOldTick = nTick;
		}

		if ( bAutoBaudDone )
		{
			bAutoBaudDone = false;

			//RS232Send_P( PSTR("Autobaud on line "));
			//RS232SendInt( nAutoBaudLine );
			//RS232Send_P( PSTR(" = "));
			long nBaud = CPU_CLK;
			nBaud /= nLastBitPeriod;
			//RS232SendLong( nBaud );
			//RS232Send_P( PSTR(" ("));
			//RS232SendLong( nLastBitPeriod );
			//RS232Send_P( PSTR(")\r\n"));
			cli();
			RS232SendChar( TYPE_AUTOBAUD | (nAutoBaudLine << 1) );
			RS232SendChar( *((byte *)&nBaud+0) );
			RS232SendChar( *((byte *)&nBaud+1) );
			RS232SendChar( *((byte *)&nBaud+2) );
			RS232SendChar( *((byte *)&nBaud+3) );
			sei();
			Line[nAutoBaudLine].Baud = nBaud;
			SetPrescaler( nAutoBaudLine, nLastBitPeriod );
			SetClockTimer();
			//StartListening();
		}

		if ( RS232ReadDataAvailable() )
		{
			ProcessComms();
		}
	}
}

/*


	ext int starts bit
		timer then measures bits.

	sec timer -> 
		count spaces?

	timer set for bytes
		baud/(start + data + stop)
		count interrupts
		if count == 255,
			send 255.
		if int
			send count


Inputs
	Config Line
		Baud
		Data Bits	5,6,7,8,9
		Parity		odd, even, none
		Stop bits	1,1.5,2

		XX - Cmd, 0x02 = config line 1, 0x03 = config line 2
		XX XX XX XX - Baud
		XX - Data bits 5, 6, 7, 8, 9
		XX - parity, 0=none, 1=odd, 2=even
		XX - stop bits 0=1, 1=1.5, 2=2

	Auto Config Line
		XX - Cmd, 0x04 = auto config line 1, 0x05 = auto config line 2
	Start
		XX - Cmd, 0x06 - start
	Stop
		XX - Cmd, 0x07 - stop

 
 * Data
    00000001 XXXXXXXX       Line 1
    00000000 XXXXXXXX       Line 2
 * Error
    00000011 XXXXXXXX       Line 1
    00000010 XXXXXXXX       Line 2


*/

